/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.5 Funktionen Ganglinie
 * Copyright (C) 2011-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.gllib.gllib.portals;

import java.util.ArrayList;
import java.util.List;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.Data.Array;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.iav.gllib.gllib.domain.ApproximationTyp;
import de.bsvrz.iav.gllib.gllib.domain.Intervall;
import de.bsvrz.iav.gllib.gllib.domain.MessQuerschnitt;
import de.bsvrz.iav.gllib.gllib.domain.Messwerte;
import de.bsvrz.iav.gllib.gllib.domain.MqGanglinie;
import de.bsvrz.iav.gllib.gllib.domain.comm.GlProgAntwort;
import de.bsvrz.iav.gllib.gllib.shared.Event;

/**
 * 
 * Implementierung des Empfängers zum Empfang von Antworten auf Anfragen an die
 * Ganglinienprognose über die Datenverteilerschnittstelle.
 * 
 * @author BitCtrl Systems GmbH, Uwe Peuker
 */
public class DavPrognoseAntwortEmpfaenger {

	private final ClientReceiverInterface receiver = new ClientReceiverInterface() {

		@Override
		public void update(ResultData[] results) {
			for (ResultData result : results) {
				if (result.hasData()) {
					empfangenEvent.send(konvertiere(result.getData()));
				}
			}
		}
	};

	private final Event<GlProgAntwort> empfangenEvent = new Event<>();
	private final DavObjektFactory objektFactory;
	private final ClientDavInterface dav;

	public DavPrognoseAntwortEmpfaenger(ClientDavInterface dav, DavObjektFactory objektFactory) {

		this.dav = dav;
		this.objektFactory = objektFactory;

		AttributeGroup atg = dav.getDataModel().getAttributeGroup("atg.prognoseGanglinienAntwort");
		Aspect asp = dav.getDataModel().getAspect("asp.antwort");
		dav.subscribeReceiver(receiver, dav.getLocalApplicationObject(), new DataDescription(atg, asp),
				ReceiveOptions.normal(), ReceiverRole.drain());
	}

	private GlProgAntwort konvertiere(Data daten) {

		String absenderzeichen = daten.getTextValue("AbsenderZeichen").getText();
		List<MqGanglinie> prognosen = new ArrayList<>();

		Array ganglinienArray = daten.getArray("PrognoseGanglinie");
		for (int i = 0; i < ganglinienArray.getLength(); ++i) {

			Data glItem = ganglinienArray.getItem(i);

			SystemObject mqObject = glItem.getReferenceValue("Messquerschnitt").getSystemObject();
			MessQuerschnitt mq = objektFactory.createMessQuerschnitt(mqObject);

			MqGanglinie g = new MqGanglinie(mq);

			// Verfahren
			g.setApproximationDaK(ApproximationTyp.with(glItem.getUnscaledValue("GanglinienVerfahren").intValue()));
			g.setBSplineOrdnung((byte) glItem.getUnscaledValue("Ordnung").longValue());

			// Prognosezeitraum
			Intervall prognoseZeitraum = Intervall.of(glItem.getTimeValue("ZeitpunktPrognoseBeginn").getMillis(),
					glItem.getTimeValue("ZeitpunktPrognoseEnde").getMillis());
			g.setPrognoseZeitraum(prognoseZeitraum);

			// Stützstellen
			Array stuetzStellenArray = glItem.getArray("Stützstelle");
			for (int j = 0; j < stuetzStellenArray.getLength(); j++) {

				Data stuetzstellenItem = stuetzStellenArray.getItem(j);
				long zeitstempel = stuetzstellenItem.getTimeValue("Zeit").getMillis();

				Double qKfz0 = null;
				Double qLkw0 = null;
				Double vPkw0 = null;
				Double vLkw0 = null;

				if (Double.compare(stuetzstellenItem.getScaledValue("QKfz").doubleValue(),
						Messwerte.UNDEFINIERT) != 0) {
					qKfz0 = stuetzstellenItem.getScaledValue("QKfz").doubleValue();
				}
				if (Double.compare(stuetzstellenItem.getScaledValue("QLkw").doubleValue(),
						Messwerte.UNDEFINIERT) != 0) {
					qLkw0 = stuetzstellenItem.getScaledValue("QLkw").doubleValue();
				}
				if (Double.compare(stuetzstellenItem.getScaledValue("VPkw").doubleValue(),
						Messwerte.UNDEFINIERT) != 0) {
					vPkw0 = stuetzstellenItem.getScaledValue("VPkw").doubleValue();
				}
				if (Double.compare(stuetzstellenItem.getScaledValue("VLkw").doubleValue(),
						Messwerte.UNDEFINIERT) != 0) {
					vLkw0 = stuetzstellenItem.getScaledValue("VLkw").doubleValue();
				}
				g.aktualisiereMesswert(zeitstempel, new Messwerte(qKfz0, qLkw0, vPkw0, vLkw0));
			}

			prognosen.add(g);
		}

		return new GlProgAntwort(objektFactory.createAnfrager(dav.getLocalApplicationObject().getId()), absenderzeichen,
				prognosen);
	}

	public Event<GlProgAntwort> antwortEmpfangen() {
		return empfangenEvent;
	}

}
