/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.5 Funktionen Ganglinie
 * Copyright (C) 2011-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.gllib.gllib.domain;

import java.text.MessageFormat;
import java.time.Duration;
import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.SortedMap;
import java.util.TreeMap;

/**
 * Helferklasse mit den Operationen auf Messquerschnittsganglinien.
 * <p>
 * <em>Hinweis:</em> Die Ergebnisganglinie bekommt den Messquerschnitt und die
 * Approximation der Ganglinie, die als erster Parameter einer Methode übergeben
 * wird. Alle anderen Eigenschaften, einer Messquerschnittsganglinie, bleiben
 * bei der Ergebnisganglinie uninitialisiert.
 *
 * @author BitCtrl Systems GmbH, Schumann
 */
public final class GanglinienMQOperationen {

	/**
	 * Kapselt die Informationen die als Ergebnis aus dem Pattern-Matching
	 * hervorgehen.
	 */
	public static final class MqGangliniePatternMatchingErgebnis {

		/** Die Ergebnisganglinie. */
		private final MqGanglinie ganglinie;

		/** Der Index der ausgewählten Ganglinie in der Vergleichsliste. */
		private final int index;

		/** Der Abstand der Ergebnisganglinie zur Referenzganglinie. */
		private final double abstand;

		/** Der Offset um den die Ergebnisganglinie verschoben wurde. */
		private final long offset;

		/**
		 * Erzeugt ein Pattern-Matching-Ergebnis.
		 *
		 * @param ganglinie die Ergebnisganglinie.
		 * @param index     der Index der ausgewählten Ganglinie in der Vergleichsliste.
		 * @param abstand   der Abstand der Ergebnisganglinie zur Referenzganglinie.
		 * @param offset    der Offset um den die Ergebnisganglinie verschoben wurde.
		 */
		public MqGangliniePatternMatchingErgebnis(final MqGanglinie ganglinie, final int index, final double abstand,
				final long offset) {
			this.ganglinie = ganglinie;
			this.index = index;
			this.abstand = abstand;
			this.offset = offset;
		}

		/**
		 * Gibt den Abstand der Ergebnisganglinie zur Referenzganglinie zurück.
		 *
		 * @return der Abstand.
		 */
		public double getAbstand() {
			return abstand;
		}

		/**
		 * Gibt die Ergebnisganglinie zurück.
		 *
		 * @return die Ergebnisganglinie.
		 */
		public MqGanglinie getGanglinie() {
			return ganglinie.erzeugeKopie();
		}

		/**
		 * Gibt den Index der ausgewählten Ganglinie in der Vergleichsliste zurück.
		 *
		 * @return der Index.
		 */
		public int getIndex() {
			return index;
		}

		/**
		 *
		 * Gibt der Offset um den die Ergebnisganglinie verschoben wurde zurück.
		 *
		 * @return der Offset der Ergebnisganglinie.
		 */
		public long getOffset() {
			return offset;
		}

		@Override
		public String toString() {
			String s;

			s = getClass() + "[";
			s += "index=" + index;
			s += ", abstand=" + abstand;
			s += ", offset=" + Duration.ofMillis(offset);
			s += ", ganglinie=" + ganglinie;
			s += "]";

			return s;
		}

		@Override
		public int hashCode() {
			return Objects.hash(abstand, ganglinie, index, offset);
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj) {
				return true;
			}
			if (obj == null) {
				return false;
			}
			if (!(obj instanceof MqGangliniePatternMatchingErgebnis)) {
				return false;
			}
			MqGangliniePatternMatchingErgebnis other = (MqGangliniePatternMatchingErgebnis) obj;
			return Double.doubleToLongBits(abstand) == Double.doubleToLongBits(other.abstand)
					&& Objects.equals(ganglinie, other.ganglinie) && index == other.index && offset == other.offset;
		}

	}

	private GanglinienMQOperationen() {
		// nichts
	}

	/**
	 * Konvertiert eine Ganglinie in einen lesbaren Text. Dazu werden die
	 * Stützstelle in einer einfachen Texttabelle ausgegeben. Enthält die Ganglinie
	 * keine Stützstellen wird eine kurze Notiz ausgegeben.
	 *
	 * @param g eine Ganglinie.
	 * @return ein String der jede Stützstelle auf eine eigene Zeile schreibt.
	 */
	public static String formatierterText(final MqGanglinie g) {
		StringBuilder txt = new StringBuilder();

		txt.append("Messquerschnitt: " + g.getMessQuerschnitt());
		txt.append("\nEreignistyp: " + g.getEreignisTyp());
		txt.append("\nTyp: ");
		switch (g.getTyp()) {
		case ABSOLUT:
			txt.append("absolute Ganglinie");
			break;
		case ADDITIV:
			txt.append("relative Ganglinie (additiv)");
			break;
		case MULTIPLIKATIV:
			txt.append("relative Ganglinie (multiplikativ)");
			break;
		default:
			txt.append("fehlerhafte Angabe");
			break;
		}
		txt.append("\nAnzahl Verschmelzungen: " + g.getAnzahlVerschmelzungen());
		txt.append("\nLetzte Verschmelzung: " + g.getLetzteVerschmelzung());
		txt.append("\nReferenzganglinie: " + g.isReferenz());
		txt.append("\nIntervall: " + g.getIntervall());
		txt.append("\nApproximation: ");
		switch (g.getApproximationDaK()) {
		case BSPLINE:
			txt.append("B-Spline, Ordnung " + g.getBSplineOrdnung());
			break;
		case CUBICSPLINE:
			txt.append("Cubic-Spline");
			break;
		case POLYLINE:
			txt.append("Polyline");
			break;
		case UNBESTIMMT:
			txt.append("unbestimmt");
			break;
		default:
			txt.append("fehlerhafte Angabe");
			break;
		}

		if (!g.hatMesswerte()) {
			txt.append("\nKeine Stützstellen vorhanden.");
		} else {
			txt.append("\nZeitpunkt\tQKfz\tQLkw\tVPkw\tVLkw\n");
			int count = 0;
			for (final Stuetzstelle<Messwerte> s : g.getStuetzstellen()) {
				if (++count > 100) {
					txt.append(" ... ");
					break;
				}
				txt.append(LocalDateTime.ofInstant(Instant.ofEpochMilli(s.getZeitstempel()), ZoneId.systemDefault())
						+ "\t");
				txt.append(s.getWert().getQKfz() + "\t");
				txt.append(s.getWert().getQLkw() + "\t");
				txt.append(s.getWert().getVPkw() + "\t");
				txt.append(s.getWert().getVLkw() + "\n");
			}
		}

		return txt.toString();
	}

	/**
	 * Addiert zwei Ganglinien, indem die Werte der vervollständigten
	 * Stützstellenmenge addiert werden. Die beiden Ganglinien werden dabei nicht
	 * verändert. Die Metainformationen der ersten Ganglinien werden in die
	 * Ergebnisganglinie kopiert.
	 *
	 * @param g1 Erste Ganglinie
	 * @param g2 Zweite Ganglinie
	 * @return Die "Summe" der beiden Ganglinien
	 */
	public static MqGanglinie addiere(final MqGanglinie g1, final MqGanglinie g2) {
		final DoubleGanglinie gQKfz;
		final DoubleGanglinie gQLkw;
		final DoubleGanglinie gVPkw;
		final DoubleGanglinie gVLkw;

		gQKfz = GanglinienOperationen.addiere(g1.getGanglinieQKfz(), g2.getGanglinieQKfz());
		gQLkw = GanglinienOperationen.addiere(g1.getGanglinieQLkw(), g2.getGanglinieQLkw());
		gVPkw = GanglinienOperationen.addiere(g1.getGanglinieVPkw(), g2.getGanglinieVPkw());
		gVLkw = GanglinienOperationen.addiere(g1.getGanglinieVLkw(), g2.getGanglinieVLkw());

		MqGanglinie mqGanglinie = new MqGanglinie(0L, g1.getMessQuerschnitt(), g1.getEreignisTyp());
		kopiereMetaDaten(mqGanglinie, g1);
		mqGanglinie.aktualisiereMesswerte(zusammenfuehren(gQKfz, gQLkw, gVPkw, gVLkw));
		return mqGanglinie;
	}

	/**
	 * Schneidet ein Intervall aus einer MqGanglinie heraus. Existieren keine
	 * Stützstellen in den Intervallgrenzen, werden an diesen Stellen mittels
	 * Approximation durch Polyline Stützstellen hinzugefügt.
	 * <p>
	 * <em>Hinweis:</em> Das Intervall wird aus der Ganglinie im Parameter
	 * ausgeschnitten.
	 *
	 * @param g Eine Ganglinie
	 * @param i Auszuschneidendes Intervall
	 * @return Der Intervallausschnitt
	 */
	public static MqGanglinie auschneiden(final MqGanglinie g, final Intervall i) {
		final DoubleGanglinie gQKfz;
		final DoubleGanglinie gQLkw;
		final DoubleGanglinie gVPkw;
		final DoubleGanglinie gVLkw;

		gQKfz = GanglinienOperationen.auschneiden(g.getGanglinieQKfz(), i);
		gQLkw = GanglinienOperationen.auschneiden(g.getGanglinieQLkw(), i);
		gVPkw = GanglinienOperationen.auschneiden(g.getGanglinieVPkw(), i);
		gVLkw = GanglinienOperationen.auschneiden(g.getGanglinieVLkw(), i);

		g.loescheMesswerte();
		g.aktualisiereMesswerte(zusammenfuehren(gQKfz, gQLkw, gVPkw, gVLkw));
		return g;
	}

	/**
	 * Berechnet den Abstand zweier Ganglinien mit Hilfe des Basisabstandsverfahren.
	 *
	 * @param g1 Erste Ganglinie
	 * @param g2 Zweite Ganglinie
	 * @return Abstand nach dem Basisabstandsverfahren
	 */
	public static double basisabstand(final MqGanglinie g1, final MqGanglinie g2) {
		double fehlerQKfz, fehlerQLkw, fehlerVPkw, fehlerVLkw;

		fehlerQKfz = GanglinienOperationen.basisabstand(g1.getGanglinieQKfz(), g2.getGanglinieQKfz());
		fehlerQLkw = GanglinienOperationen.basisabstand(g1.getGanglinieQLkw(), g2.getGanglinieQLkw());
		fehlerVPkw = GanglinienOperationen.basisabstand(g1.getGanglinieVPkw(), g2.getGanglinieVPkw());
		fehlerVLkw = GanglinienOperationen.basisabstand(g1.getGanglinieVLkw(), g2.getGanglinieVLkw());

		return ((fehlerQKfz + fehlerQLkw + fehlerVLkw + fehlerVPkw) / 4);
	}

	/**
	 * Division zweier Ganglinien, indem die Werte der vervollständigten
	 * Stützstellenmenge dividiert werden. Die beiden Ganglinien werden dabei nicht
	 * verändert. Die Metainformationen der ersten Ganglinien werden in die
	 * Ergebnisganglinie kopiert.
	 *
	 * @param g1 Erste Ganglinie
	 * @param g2 Zweite Ganglinie
	 * @return Das "Produkt" der beiden Ganglinien
	 */
	public static MqGanglinie dividiere(final MqGanglinie g1, final MqGanglinie g2) {
		final DoubleGanglinie gQKfz;
		final DoubleGanglinie gQLkw;
		final DoubleGanglinie gVPkw;
		final DoubleGanglinie gVLkw;

		assert g1.getMessQuerschnitt()
				.equals(g2.getMessQuerschnitt()) : "Die Ganglinien müssen zum gleichen Messquerschnitt gehören.";

		gQKfz = GanglinienOperationen.dividiere(g1.getGanglinieQKfz(), g2.getGanglinieQKfz());
		gQLkw = GanglinienOperationen.dividiere(g1.getGanglinieQLkw(), g2.getGanglinieQLkw());
		gVPkw = GanglinienOperationen.dividiere(g1.getGanglinieVPkw(), g2.getGanglinieVPkw());
		gVLkw = GanglinienOperationen.dividiere(g1.getGanglinieVLkw(), g2.getGanglinieVLkw());

		MqGanglinie mqGanglinie = new MqGanglinie(0L, g1.getMessQuerschnitt(), g1.getEreignisTyp());
		kopiereMetaDaten(mqGanglinie, g1);
		mqGanglinie.aktualisiereMesswerte(zusammenfuehren(gQKfz, gQLkw, gVPkw, gVLkw));
		return mqGanglinie;
	}

	/**
	 * Berechnet den Abstand zweier Ganglinien mit Hilfe des komplexen
	 * Abstandsverfahren. Es werden jeweils die Werte an den Intervallgrenzen
	 * verglichen.
	 *
	 * @param g1         Erste Ganglinie
	 * @param g2         Zweite Ganglinie
	 * @param intervalle Anzahl der zu vergleichenden Intervalle
	 * @return Abstand nach dem komplexen Abstandsverfahren
	 */
	public static double komplexerAbstand(final MqGanglinie g1, final MqGanglinie g2, final int intervalle) {
		assert g1.getMessQuerschnitt()
				.equals(g2.getMessQuerschnitt()) : "Die Ganglinien müssen zum gleichen Messquerschnitt gehören.";

		double fehlerQKfz, fehlerQLkw, fehlerVPkw, fehlerVLkw;

		fehlerQKfz = GanglinienOperationen.komplexerAbstand(g1.getGanglinieQKfz(), g2.getGanglinieQKfz(), intervalle);
		fehlerQLkw = GanglinienOperationen.komplexerAbstand(g1.getGanglinieQLkw(), g2.getGanglinieQLkw(), intervalle);
		fehlerVPkw = GanglinienOperationen.komplexerAbstand(g1.getGanglinieVPkw(), g2.getGanglinieVPkw(), intervalle);
		fehlerVLkw = GanglinienOperationen.komplexerAbstand(g1.getGanglinieVLkw(), g2.getGanglinieVLkw(), intervalle);

		return (fehlerQKfz + fehlerQLkw + fehlerVLkw + fehlerVPkw) / 4.0;
	}

	/**
	 * Berechnet den Abstand zweier Ganglinien mit Hilfe des komplexen
	 * Abstandsverfahren. Es werden jeweils die Werte an den Intervallgrenzen
	 * verglichen.
	 *
	 * @param g1              Erste Ganglinie.
	 * @param g2              Zweite Ganglinie.
	 * @param intervallBreite die Breite der zu vergleichenden Intervalle.
	 * @return Abstand nach dem komplexen Abstandsverfahren in Prozent.
	 */
	public static double komplexerAbstand(final MqGanglinie g1, final MqGanglinie g2, final long intervallBreite) {
		assert g1.getMessQuerschnitt()
				.equals(g2.getMessQuerschnitt()) : "Die Ganglinien müssen zum gleichen Messquerschnitt gehören.";

		double fehlerQKfz, fehlerQLkw, fehlerVPkw, fehlerVLkw;

		fehlerQKfz = GanglinienOperationen.komplexerAbstand(g1.getGanglinieQKfz(), g2.getGanglinieQKfz(),
				intervallBreite);
		fehlerQLkw = GanglinienOperationen.komplexerAbstand(g1.getGanglinieQLkw(), g2.getGanglinieQLkw(),
				intervallBreite);
		fehlerVPkw = GanglinienOperationen.komplexerAbstand(g1.getGanglinieVPkw(), g2.getGanglinieVPkw(),
				intervallBreite);
		fehlerVLkw = GanglinienOperationen.komplexerAbstand(g1.getGanglinieVLkw(), g2.getGanglinieVLkw(),
				intervallBreite);

		return (fehlerQKfz + fehlerQLkw + fehlerVLkw + fehlerVPkw) / 4.0;
	}

	/**
	 * Kopiert die Metainformationen einer Ganglinie auf eine andere.
	 * <p>
	 * <em>Hinweis:</em> Es wird die Ganglinie im Parameter verändert.
	 *
	 * @param ziel   das Ziel der Metadaten.
	 * @param quelle die Quelle der Metadaten.
	 * @return die neue Ganglinie.
	 */
	private static MqGanglinie kopiereMetaDaten(final MqGanglinie ziel, final MqGanglinie quelle) {
		ziel.setAnzahlVerschmelzungen(quelle.getAnzahlVerschmelzungen());
		ziel.setApproximationDaK(quelle.getApproximationDaK());
		ziel.setBSplineOrdnung(quelle.getBSplineOrdnung());
		ziel.setK1(quelle.getK1());
		ziel.setK2(quelle.getK2());
		ziel.setLetzteVerschmelzung(quelle.getLetzteVerschmelzung());
		ziel.setPrognoseZeitraum(quelle.getPrognoseIntervall());
		ziel.setReferenz(quelle.isReferenz());
		ziel.setTyp(quelle.getTyp());

		return ziel;
	}

	/**
	 * Multiplikation zweier Ganglinien, indem die Werte der vervollständigten
	 * Stützstellenmenge multipliziert werden. Die beiden Ganglinien werden dabei
	 * nicht verändert. Die Metainformationen der ersten Ganglinien werden in die
	 * Ergebnisganglinie kopiert.
	 *
	 * @param g1 Erste Ganglinie
	 * @param g2 Zweite Ganglinie
	 * @return Das "Produkt" der beiden Ganglinien
	 */
	public static MqGanglinie multipliziere(final MqGanglinie g1, final MqGanglinie g2) {
		final DoubleGanglinie gQKfz;
		final DoubleGanglinie gQLkw;
		final DoubleGanglinie gVPkw;
		final DoubleGanglinie gVLkw;

		gQKfz = GanglinienOperationen.multipliziere(g1.getGanglinieQKfz(), g2.getGanglinieQKfz());
		gQLkw = GanglinienOperationen.multipliziere(g1.getGanglinieQLkw(), g2.getGanglinieQLkw());
		gVPkw = GanglinienOperationen.multipliziere(g1.getGanglinieVPkw(), g2.getGanglinieVPkw());
		gVLkw = GanglinienOperationen.multipliziere(g1.getGanglinieVLkw(), g2.getGanglinieVLkw());

		MqGanglinie mqGanglinie = new MqGanglinie(0L, g1.getMessQuerschnitt(), g1.getEreignisTyp());
		kopiereMetaDaten(mqGanglinie, g1);
		mqGanglinie.aktualisiereMesswerte(zusammenfuehren(gQKfz, gQLkw, gVPkw, gVLkw));
		return mqGanglinie;
	}

	/**
	 * Normiert die Stützstellen einer Ganglinie. Mehrere hintereinander folgende
	 * undefinierte Stützstellen werden zusammengefasst. Der Abstand der
	 * Stützstellen wird auf ein definiertes Intervall normiert. Der Intervallzyklus
	 * beginnt mit dem Zeitstempel der ersten Stützstelle.
	 * <p>
	 * <em>Hinweis:</em> die Ganglinie im Parameter wird verändert.
	 *
	 * @param g       eine Ganglinie.
	 * @param abstand der gewünschte Stützstellenabstand.
	 * @return die normierte Ganglinie.
	 */
	public static MqGanglinie normiere(final MqGanglinie g, final long abstand) {
		DoubleGanglinie gQKfz, gQLkw, gVPkw, gVLkw;

		gQKfz = GanglinienOperationen.normiere(g.getGanglinieQKfz(), abstand);
		gQLkw = GanglinienOperationen.normiere(g.getGanglinieQLkw(), abstand);
		gVPkw = GanglinienOperationen.normiere(g.getGanglinieVPkw(), abstand);
		gVLkw = GanglinienOperationen.normiere(g.getGanglinieVLkw(), abstand);

		g.loescheMesswerte();
		g.aktualisiereMesswerte(zusammenfuehren(gQKfz, gQLkw, gVPkw, gVLkw));
		return g;
	}

	/**
	 * Führt das Pattern-Matching einer Menge von Ganglinien mit einer
	 * Referenzganglinie aus. Das Ergebnis ist die Ganglinie aus der Menge mit dem
	 * geringsten Abstand zur Referenzganglinie.
	 * <p>
	 * Zusätzlich zu der Liste von Vergleichsganglinien wird jede dieser Ganglinien
	 * im angegebenen Offset in {@code intervall} Schritten verschoben. Jede dieser
	 * so entstanden Ganglinien wird ebenfalls mit der Referenzganglinie verglichen.
	 * <p>
	 * Ergebnis ist die vorgegebene oder erzeugte Ganglinie mit dem geringsten
	 * Abstand. Der Index im Ergebnis ist der Index in der Ganglinienliste, aus der
	 * die Ergebnisganglinie hervorgegangen ist.
	 *
	 * @param referenz        die Referenzganglinie.
	 * @param vergleichsListe die Liste von zu vergleichenden Ganglinien.
	 * @param offsetVor       der Offset, in dem die Ganglinien nach vorn verschoben
	 *                        werden kann.
	 * @param offsetNach      der Offset, in dem die Ganglinien nach hinten
	 *                        verschoben werden kann.
	 * @param schrittweite    das Intervall, in dem die Ganglinien innerhalb des
	 *                        Offsets verschoben werden.
	 * @return das Ergebnis des Pattern-Matching.
	 */
	public static GanglinienMQOperationen.MqGangliniePatternMatchingErgebnis patternMatching(final MqGanglinie referenz,
			final List<MqGanglinie> vergleichsListe, final long offsetVor, final long offsetNach,
			final long schrittweite) {
		if (referenz == null) {
			throw new IllegalArgumentException(MessageFormat.format("Argument darf nicht null sein: {0}", "referenz"));
		}
		if (vergleichsListe.isEmpty()) {
			throw new IllegalArgumentException(
					MessageFormat.format("Argument darf keine leere Liste sein: {0}", "vergleichsListe"));
		}
		if (offsetNach < 0) {
			throw new IllegalArgumentException(MessageFormat.format(
					"IllegalArgumentExceptionGroesserGleichNull = Argument muss eine Zahl größer oder gleich 0 sein: {0}",
					"offsetNach"));
		}
		if (offsetVor < 0) {
			throw new IllegalArgumentException(MessageFormat.format(
					"IllegalArgumentExceptionGroesserGleichNull = Argument muss eine Zahl größer oder gleich 0 sein: {0}",
					"offsetVor"));
		}
		if (schrittweite <= 0) {
			throw new IllegalArgumentException(
					MessageFormat.format("Argument muss eine Zahl größer 0 sein: {0}", "schrittweite"));
		}

		MqGangliniePatternMatchingErgebnis ergebnis = null;
		for (int i = 0; i < vergleichsListe.size(); i++) {
			for (long offset = -offsetVor; offset <= offsetNach; offset += schrittweite) {
				MqGanglinie g = vergleichsListe.get(i).erzeugeKopie();
				verschiebe(g, offset);
				double abstand = basisabstand(referenz, g);
				MqGangliniePatternMatchingErgebnis matchingErgebnis = new MqGangliniePatternMatchingErgebnis(
						g.erzeugeKopie(), i, abstand, offset);
				if (ergebnis == null || ergebnis.abstand > matchingErgebnis.abstand) {
					ergebnis = matchingErgebnis;
				}
			}
		}

		return ergebnis;
	}

	/**
	 * Subtraktion zweier Ganglinien, indem die Werte der vervollständigten
	 * Stützstellenmenge subtrahiert werden. Die beiden Ganglinien werden dabei
	 * nicht verändert. Die Metainformationen der ersten Ganglinien werden in die
	 * Ergebnisganglinie kopiert.
	 *
	 * @param g1 Erste Ganglinie
	 * @param g2 Zweite Ganglinie
	 * @return Die "Differenz" der beiden Ganglinien
	 */
	public static MqGanglinie subtrahiere(final MqGanglinie g1, final MqGanglinie g2) {
		final DoubleGanglinie gQKfz;
		final DoubleGanglinie gQLkw;
		final DoubleGanglinie gVPkw;
		final DoubleGanglinie gVLkw;

		gQKfz = GanglinienOperationen.subtrahiere(g1.getGanglinieQKfz(), g2.getGanglinieQKfz());
		gQLkw = GanglinienOperationen.subtrahiere(g1.getGanglinieQLkw(), g2.getGanglinieQLkw());
		gVPkw = GanglinienOperationen.subtrahiere(g1.getGanglinieVPkw(), g2.getGanglinieVPkw());
		gVLkw = GanglinienOperationen.subtrahiere(g1.getGanglinieVLkw(), g2.getGanglinieVLkw());

		MqGanglinie mqGanglinie = new MqGanglinie(0L, g1.getMessQuerschnitt(), g1.getEreignisTyp());
		kopiereMetaDaten(mqGanglinie, g1);
		mqGanglinie.aktualisiereMesswerte(zusammenfuehren(gQKfz, gQLkw, gVPkw, gVLkw));
		return mqGanglinie;
	}

	/**
	 * Verbindet zwei Ganglinien durch Konkatenation. Es werden die Stützstellen
	 * beider Ganglinien zu einer neuen Ganglinien zusammengefasst. Dies ist nur
	 * möglich, wenn sich die Stützstellenmengen nicht überschneiden. Berühren sich
	 * die beiden Ganglinien wird im Berührungspunkt er Mittelwert der beiden
	 * Stützstellen gebildet. Die Metainformationen der ersten Ganglinien werden in
	 * die Ergebnisganglinie kopiert.
	 * <p>
	 * Wird der maximale Abstand der beiden Ganglinien überschritten, ist der
	 * Bereich zwischen den Ganglinien undefiniert.
	 *
	 * @param g1         Erste Ganglinie
	 * @param g2         Zweite Ganglinie
	 * @param maxAbstand der maximale Abstand der beiden Ganglinien.
	 * @return Konkatenation der beiden Ganglinien
	 */
	public static MqGanglinie verbinde(final MqGanglinie g1, final MqGanglinie g2, final long maxAbstand) {
		final DoubleGanglinie gQKfz, gQLkw, gVPkw, gVLkw;

		gQKfz = GanglinienOperationen.verbinde(g1.getGanglinieQKfz(), g2.getGanglinieQKfz(), maxAbstand);
		gQLkw = GanglinienOperationen.verbinde(g1.getGanglinieQLkw(), g2.getGanglinieQLkw(), maxAbstand);
		gVPkw = GanglinienOperationen.verbinde(g1.getGanglinieVPkw(), g2.getGanglinieVPkw(), maxAbstand);
		gVLkw = GanglinienOperationen.verbinde(g1.getGanglinieVLkw(), g2.getGanglinieVLkw(), maxAbstand);

		MqGanglinie mqGanglinie = new MqGanglinie(0L, g1.getMessQuerschnitt(), g1.getEreignisTyp());
		kopiereMetaDaten(mqGanglinie, g1);
		mqGanglinie.aktualisiereMesswerte(zusammenfuehren(gQKfz, gQLkw, gVPkw, gVLkw));
		return mqGanglinie;
	}

	/**
	 * Verschiebt eine MqGanglinie auf der Zeitachse.
	 * <p>
	 * <em>Hinweis:</em> Es wird die Ganglinie im Parameter verschoben.
	 *
	 * @param g      Zu verschiebende Ganglinie
	 * @param offset Offset um den die MqGanglinie verschoben werden soll
	 * @return Die verschobene Ganglinie
	 */
	public static MqGanglinie verschiebe(final MqGanglinie g, final long offset) {
		DoubleGanglinie gQKfz, gQLkw, gVPkw, gVLkw;

		gQKfz = GanglinienOperationen.verschiebe(g.getGanglinieQKfz(), offset);
		gQLkw = GanglinienOperationen.verschiebe(g.getGanglinieQLkw(), offset);
		gVPkw = GanglinienOperationen.verschiebe(g.getGanglinieVPkw(), offset);
		gVLkw = GanglinienOperationen.verschiebe(g.getGanglinieVLkw(), offset);

		g.loescheMesswerte();
//        Intervall vorherigesIntervall = g.getPrognoseIntervall();
//        if (vorherigesIntervall != null) {
//            g.setPrognoseZeitraum(
//                    Intervall.of(vorherigesIntervall.getStart() + offset, vorherigesIntervall.getEnd() + offset));
//        }
		g.aktualisiereMesswerte(zusammenfuehren(gQKfz, gQLkw, gVPkw, gVLkw));
		return g;
	}

	/**
	 * Verschiebt eine Ganglinie auf der Zeitachse um ein halbes
	 * Stützstellenintervall. Jede Stützstelle wird um den halben Abstand zur
	 * nächsten Stützstelle verschoben. Die letzte Stützstelle wird um den halben
	 * Abstand zur vorherigen Stützstelle verschoben. Gibt es nur eine Stützstelle,
	 * wird diese nicht verschoben.
	 * <p>
	 * <em>Hinweis:</em> Es wird die Ganglinie im Parameter verschoben.
	 *
	 * @param g Zu verschiebende Ganglinie
	 * @return Die verschobene Ganglinie
	 */
	public static MqGanglinie verschiebeUmHalbesIntervall(final MqGanglinie g) {
		DoubleGanglinie gQKfz, gQLkw, gVPkw, gVLkw;

		gQKfz = GanglinienOperationen.verschiebeUmHalbesIntervall(g.getGanglinieQKfz());
		gQLkw = GanglinienOperationen.verschiebeUmHalbesIntervall(g.getGanglinieQLkw());
		gVPkw = GanglinienOperationen.verschiebeUmHalbesIntervall(g.getGanglinieVPkw());
		gVLkw = GanglinienOperationen.verschiebeUmHalbesIntervall(g.getGanglinieVLkw());

		g.loescheMesswerte();
		g.aktualisiereMesswerte(zusammenfuehren(gQKfz, gQLkw, gVPkw, gVLkw));
		return g;
	}

	/**
	 * Verschmilzt eine Ganglinie mit einer anderen. Dabei wird das gewichtete
	 * arithmetische Mittel der vervollständigten Stützstellen gebildet. Die zweite
	 * Ganglinie hat immer das Gewicht 1. Die beiden Ganglinien werden dabei nicht
	 * verändert.
	 * <p>
	 * Die Anzahl der Verschmelzungen und der Zeitpunkt der letzten Verschmelzungen
	 * werden aktualisiert.
	 * <p>
	 * <em>Hinweis:</em> die historische Ganglinie im Parameter wird geändert.
	 *
	 * @param verschmelzungsZeit der Zeitpunkt für den die Verschmelzung erfolgen
	 *                           soll
	 * @param ganglinie          die Ganglinie mit der verschmolzen wird. Sie hat
	 *                           immer das Gewicht 1.
	 * @param historGl           die historische Ganglinie die verschmolzen wird.
	 * @param gewicht            das Gewicht der zweiten Ganglinie.
	 * @return das Ergebnis der Verschmelzung.
	 */
	public static MqGanglinie verschmelze(LocalDateTime verschmelzungsZeit, final MqGanglinie ganglinie,
			final MqGanglinie historGl, final long gewicht) {

		final DoubleGanglinie gQKfz, gQLkw, gVPkw, gVLkw;

		gQKfz = GanglinienOperationen.verschmelze(ganglinie.getGanglinieQKfz(), historGl.getGanglinieQKfz(), gewicht);
		gQLkw = GanglinienOperationen.verschmelze(ganglinie.getGanglinieQLkw(), historGl.getGanglinieQLkw(), gewicht);
		gVPkw = GanglinienOperationen.verschmelze(ganglinie.getGanglinieVPkw(), historGl.getGanglinieVPkw(), gewicht);
		gVLkw = GanglinienOperationen.verschmelze(ganglinie.getGanglinieVLkw(), historGl.getGanglinieVLkw(), gewicht);

		historGl.loescheMesswerte();
		historGl.aktualisiereMesswerte(zusammenfuehren(gQKfz, gQLkw, gVPkw, gVLkw));
		historGl.setAnzahlVerschmelzungen(historGl.getAnzahlVerschmelzungen() + 1);
		historGl.setLetzteVerschmelzung(verschmelzungsZeit);

		return historGl;
	}

	/**
	 * Erzeugt aus den vier Ganglinien eine Messquerschnittsganglinie.
	 *
	 * @param gQKfz die Ganglinie für QKfz.
	 * @param gQLkw die Ganglinie für QLkw.
	 * @param gVPkw die Ganglinie für VPkw.
	 * @param gVLkw die Ganglinie für VLkw.
	 * @return die zusammengeführte Ganglinie.
	 */
	public static Map<Long, Messwerte> zusammenfuehren(final DoubleGanglinie gQKfz, final DoubleGanglinie gQLkw,
			final DoubleGanglinie gVPkw, final DoubleGanglinie gVLkw) {

		assert (gQKfz.punkteAnzahl() == gQLkw.punkteAnzahl()) && (gQLkw.punkteAnzahl() == gVPkw.punkteAnzahl())
				&& (gVPkw.punkteAnzahl() == gVLkw
						.punkteAnzahl()) : "Die berechneten Stützstellenlisten müssen gleich groß sein.";

		final SortedMap<Long, Messwerte> stuetzstellen = new TreeMap<>();
		for (final Long t : gQKfz.getZeitpunkte()) {
			assert gQKfz.hatZeitpunkt(t) && gQLkw.hatZeitpunkt(t) && gVPkw.hatZeitpunkt(t)
					&& gVLkw.hatZeitpunkt(t) : "Für jede Ganglinie muss für jeden Zeitstempel ein Wert existieren.";
			stuetzstellen.put(t, new Messwerte(gQKfz.getWert(t), gQLkw.getWert(t), gVPkw.getWert(t), gVLkw.getWert(t)));
		}

		return stuetzstellen;
	}

}
