/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.5 Funktionen Ganglinie
 * Copyright (C) 2011-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.gllib.gllib.domain;

import java.util.List;

/**
 * Approximation einer Ganglinie mit Hilfe von Polylines. Der Wert der
 * Stützstelle zu einem Zeitstempel wird nach folgender Formel berechnet:
 * <p>
 * <img src="doc-files/formel_polyline.png" alt="Formel für Polylinie">
 *
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public class Polyline extends AbstractApproximation<Double> {

    @Override
    public Stuetzstelle<Double> get(final long zeitstempel) {
        Stuetzstelle<Double> s0, s1;
        Double wert;
        double x0, x1, y0, y1;
        int index;

        List<Stuetzstelle<Double>> stuetzstellen = getStuetzstellen();
        if ((stuetzstellen.size() == 0) || ((stuetzstellen.get(0).getZeitstempel() < zeitstempel)
                && (zeitstempel > stuetzstellen.get(stuetzstellen.size() - 1).getZeitstempel()))) {
            // Zeitstempel liegt außerhalb der Ganglinie
            return new Stuetzstelle<>(zeitstempel, null);
        }

        index = findeStuetzstelleVor(zeitstempel);
        if (index == -1) {
            // Zeitstempel liegt außerhalb der Ganglinie
            return new Stuetzstelle<>(zeitstempel, null);
        }

        if (stuetzstellen.get(index).getZeitstempel() == zeitstempel) {
            return stuetzstellen.get(index);
        }

        s0 = stuetzstellen.get(index);
        s1 = stuetzstellen.get(index + 1);
        x0 = s0.getZeitstempel();
        y0 = s0.getWert().doubleValue();
        x1 = s1.getZeitstempel();
        y1 = s1.getWert().doubleValue();
        wert = y0 + (((y1 - y0) / (x1 - x0)) * (zeitstempel - x0));
        return new Stuetzstelle<>(zeitstempel, wert);
    }

    @Override
    public void initialisiere() {
        // Hier gibt es nichts zu tun.
    }

    /**
     * Die festgelegte Breite der Teilintervalle wird ignoriert, da sich das
     * Integral der Polylinie exakt bestimmen lässt.
     *
     * {@inheritDoc}
     */
    @Override
    public double integral(final Intervall intervall) {
        final int start;
        final int ende;
        double flaeche = 0;

        start = findeStuetzstelleVor(intervall.getStart());
        ende = findeStuetzstelleNach(intervall.getEnd());

        List<Stuetzstelle<Double>> stuetzstellen = getStuetzstellen();
        for (int i = start; i < ende; ++i) {
            long breite;
            int iRechteck, iDreieck;

            breite = stuetzstellen.get(i + 1).getZeitstempel() - stuetzstellen.get(i).getZeitstempel();
            if (stuetzstellen.get(i).getWert() < stuetzstellen.get(i + 1).getWert()) {
                iRechteck = i;
                iDreieck = i + 1;
            } else {
                iRechteck = i + 1;
                iDreieck = i;
            }

            flaeche += stuetzstellen.get(iRechteck).getWert() * breite;
            flaeche += ((stuetzstellen.get(iDreieck).getWert() - stuetzstellen.get(iRechteck).getWert()) * breite) / 2;
        }

        if (stuetzstellen.get(start).getZeitstempel() < intervall.getStart()) {
            // Erste Stützstelle liegt vor Intervall
            long breite;
            double hDreieck, hRechteck;

            hRechteck = stuetzstellen.get(start).getWert();
            hDreieck = get(intervall.getStart()).getWert();
            if (hRechteck > hDreieck) {
                // Das Rechteck muss die kleinere Höhe haben.
                double tmp;

                tmp = hRechteck;
                hRechteck = hDreieck;
                hDreieck = tmp;
            }

            breite = intervall.getStart() - stuetzstellen.get(start).getZeitstempel();
            flaeche -= hRechteck * breite;
            flaeche -= ((hDreieck - hRechteck) * breite) / 2;
        }

        if (stuetzstellen.get(ende).getZeitstempel() > intervall.getEnd()) {
            // Letzte Stützstelle liegt vor Intervall
            long breite;
            double hDreieck, hRechteck;

            hRechteck = stuetzstellen.get(ende).getWert();
            hDreieck = get(intervall.getEnd()).getWert();
            if (hRechteck > hDreieck) {
                // Das Rechteck muss die kleinere Höhe haben.
                double tmp;

                tmp = hRechteck;
                hRechteck = hDreieck;
                hDreieck = tmp;
            }

            breite = stuetzstellen.get(ende).getZeitstempel() - intervall.getEnd();
            flaeche -= hRechteck * breite;
            flaeche -= ((hDreieck - hRechteck) * breite) / 2;
        }

        return flaeche;
    }

    @Override
    public String toString() {
        return "Polylinie";
    }

}
