/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.5 Funktionen Ganglinie
 * Copyright (C) 2011-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.gllib.gllib.portals;

import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.util.Objects;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.Data.Array;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.iav.gllib.gllib.domain.ApproximationTyp;
import de.bsvrz.iav.gllib.gllib.domain.EreignisTyp;
import de.bsvrz.iav.gllib.gllib.domain.GanglinieAnfrageTyp;
import de.bsvrz.iav.gllib.gllib.domain.GanglinieTyp;
import de.bsvrz.iav.gllib.gllib.domain.MessQuerschnitt;
import de.bsvrz.iav.gllib.gllib.domain.Messwerte;
import de.bsvrz.iav.gllib.gllib.domain.MqGanglinie;
import de.bsvrz.iav.gllib.gllib.domain.comm.GlSpeicherAntwort;
import de.bsvrz.iav.gllib.gllib.domain.comm.GlSpeicherFehlerAntwort;
import de.bsvrz.iav.gllib.gllib.shared.Event;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * 
 * Implementierung eines Empfängers zum Empfangen von Antworten auf Anfragen an
 * den Ganglinienspeicher über die Datenverteilerschnittstelle.
 * 
 * @author BitCtrl Systems GmbH, Uwe Peuker
 */
public class DavGanglinieAntwortEmpfaenger {

    private static final Debug LOGGER = Debug.getLogger();

    private final ClientReceiverInterface receiver = new ClientReceiverInterface() {

        @Override
        public void update(ResultData[] results) {
            for (ResultData result : results) {
                if (result.hasData()) {
                    LOGGER.finer(() -> "Verteile Antwort:  " + result);
                    verteileAntwort(result.getData());
                }
            }
        }
    };

    private final Event<GlSpeicherAntwort> antwortEmpfangen = new Event<>();
    private final Event<GlSpeicherFehlerAntwort> fehlerAntwortEmpfangen = new Event<>();

    private ClientDavInterface dav;
    private DavObjektFactory objektFactory;

    public DavGanglinieAntwortEmpfaenger(ClientDavInterface dav, DavObjektFactory objektFactory,
            SystemObject speicher) {
        this.dav = dav;
        this.objektFactory = objektFactory;
        AttributeGroup atg = dav.getDataModel().getAttributeGroup("atg.ganglinienSpeicherAntwort");
        Aspect asp = dav.getDataModel().getAspect("asp.antwort");
        dav.subscribeReceiver(receiver, speicher, new DataDescription(atg, asp), ReceiveOptions.normal(),
                ReceiverRole.receiver());
    }

    public Event<GlSpeicherAntwort> antwortEmpfangen() {
        return antwortEmpfangen;
    }

    public Event<GlSpeicherFehlerAntwort> fehlerAntwortEmpfangen() {
        return fehlerAntwortEmpfangen;
    }

    private void verteileAntwort(Data data) {

		SystemObject absenderId = data.getReferenceValue("absenderId").getSystemObject();
        if (!Objects.equals(absenderId, dav.getLocalApplicationObject())) {
			return;
		}

		Array fehlerArray = data.getItem("Antwort").getArray("Fehlermeldung"); 
		if (fehlerArray.getLength() == 0) {
			antwortEmpfangen.send(erzeugeAntwort(data));
		} else {
			fehlerAntwortEmpfangen.send(erzeugeFehlerAntwort(data, fehlerArray));
		}
	}

    private GlSpeicherAntwort erzeugeAntwort(Data data) {

        SystemObject mqObj = data.getItem("Antwort").getReferenceValue("Messquerschnitt").getSystemObject();
        MessQuerschnitt mq = objektFactory.createMessQuerschnitt(mqObj);

        GlSpeicherAntwort antwort = new GlSpeicherAntwort(GanglinieAnfrageTyp.UNDEFINIERT,
                objektFactory.createAnfrager(data.getReferenceValue("absenderId").getSystemObject().getId()),
                data.getTextValue("AbsenderZeichen").getText(), mq);

        final Array davGanglinien = data.getItem("Antwort").getArray("GespeicherteGanglinien");
        for (int i = 0; i < davGanglinien.getLength(); i++) {
            Data glDaten = davGanglinien.getItem(i);
            antwort.addGanglinie(erzeugeMqGanglinie(mq, glDaten));
        }

        return antwort;
    }

    private GlSpeicherFehlerAntwort erzeugeFehlerAntwort(Data data, Array fehlerArray) {

        GlSpeicherFehlerAntwort antwort = new GlSpeicherFehlerAntwort(data.getTextValue("AbsenderZeichen").getText());

        for (int idx = 0; idx < fehlerArray.getLength(); idx++) {
            antwort.addFehlerMeldung(fehlerArray.getTextValue(idx).getText());
        }

        SystemObject mqObj = data.getItem("Antwort").getReferenceValue("Messquerschnitt").getSystemObject();
        if (mqObj != null) {
            MessQuerschnitt mq = objektFactory.createMessQuerschnitt(mqObj);
            antwort.setMessQuerschnitt(mq);
        }

        final Array davGanglinien = data.getItem("Antwort").getArray("GespeicherteGanglinien");
        for (int i = 0; i < davGanglinien.getLength(); i++) {
            Data glDaten = davGanglinien.getItem(i);
            antwort.addGanglinie(erzeugeMqGanglinie(antwort.getMessQuerschnitt(), glDaten));
        }

        return antwort;
    }

    private MqGanglinie erzeugeMqGanglinie(MessQuerschnitt mq, Data glDaten) {

        EreignisTyp ereignisTyp = null;
        SystemObject ereignisObj = glDaten.getReferenceValue("EreignisTyp").getSystemObject();
        if (ereignisObj != null) {
            ereignisTyp = EreignisTyp.of(ereignisObj.getPid());
        }

        long ganglinieId = glDaten.getUnscaledValue("GanglinieId").longValue();
        MqGanglinie ganglinie = new MqGanglinie(ganglinieId, mq, ereignisTyp);

        // Verfahren
        ganglinie.setBSplineOrdnung((int) glDaten.getUnscaledValue("Ordnung").longValue());
        ganglinie
                .setReferenz(glDaten.getUnscaledValue("Referenzganglinie").getState().getName().equalsIgnoreCase("ja"));
        ganglinie.setAnzahlVerschmelzungen(glDaten.getUnscaledValue("AnzahlVerschmelzungen").longValue());
        ganglinie.setLetzteVerschmelzung(LocalDateTime.ofInstant(
                Instant.ofEpochMilli(glDaten.getTimeValue("LetzteVerschmelzung").getMillis()), ZoneId.systemDefault()));
        ganglinie.setTyp(GanglinieTyp.with(glDaten.getUnscaledValue("GanglinienTyp").intValue()));
        ganglinie
                .setApproximationDaK(ApproximationTyp.with(glDaten.getUnscaledValue("GanglinienVerfahren").intValue()));

        // Stützstellen
        Array stuetzStellenArray = glDaten.getArray("Stützstelle");
        for (int j = 0; j < stuetzStellenArray.getLength(); j++) {

            Data stuetzstellenItem = stuetzStellenArray.getItem(j);
            long zeitstempel = stuetzstellenItem.getTimeValue("Zeit").getMillis();

            Double qKfz0 = null;
            Double qLkw0 = null;
            Double vPkw0 = null;
            Double vLkw0 = null;

            if (Double.compare(stuetzstellenItem.getScaledValue("QKfz").doubleValue(), Messwerte.UNDEFINIERT) != 0) {
                qKfz0 = stuetzstellenItem.getScaledValue("QKfz").doubleValue();
            }
            if (Double.compare(stuetzstellenItem.getScaledValue("QLkw").doubleValue(), Messwerte.UNDEFINIERT) != 0) {
                qLkw0 = stuetzstellenItem.getScaledValue("QLkw").doubleValue();
            }
            if (Double.compare(stuetzstellenItem.getScaledValue("VPkw").doubleValue(), Messwerte.UNDEFINIERT) != 0) {
                vPkw0 = stuetzstellenItem.getScaledValue("VPkw").doubleValue();
            }
            if (Double.compare(stuetzstellenItem.getScaledValue("VLkw").doubleValue(), Messwerte.UNDEFINIERT) != 0) {
                vLkw0 = stuetzstellenItem.getScaledValue("VLkw").doubleValue();
            }
            ganglinie.aktualisiereMesswert(zeitstempel, new Messwerte(qKfz0, qLkw0, vPkw0, vLkw0));
        }

        return ganglinie;
    }

}
